/* stylelint-disable */
/* Font Awesome Free 7.3.1 subset icon widths (Icons: CC BY 4.0, https://fontawesome.com/license/free) */
/* Auto-generated by icons-build.mjs - DO NOT EDIT. Re-run "node icons-build.mjs". */
/* Default .svg-inline--fa width is 1em (square viewBox); rules below size non-square icons. */
.svg-inline--fa[data-icon="fa-bars"] { width: 0.875em; }
.svg-inline--fa[data-icon="fa-chart-pie"] { width: 1.125em; }
.svg-inline--fa[data-icon="fa-check"] { width: 0.875em; }
.svg-inline--fa[data-icon="fa-columns"] { width: 0.875em; }
.svg-inline--fa[data-icon="fa-copy"] { width: 0.875em; }
.svg-inline--fa[data-icon="fa-eraser"] { width: 1.125em; }
.svg-inline--fa[data-icon="fa-eye"] { width: 1.125em; }
.svg-inline--fa[data-icon="fa-graduation-cap"] { width: 1.125em; }
.svg-inline--fa[data-icon="fa-save"] { width: 0.875em; }
.svg-inline--fa[data-icon="fa-server"] { width: 0.875em; }
.svg-inline--fa[data-icon="fa-sun"] { width: 1.125em; }
.svg-inline--fa[data-icon="fa-times"] { width: 0.75em; }
.svg-inline--fa[data-icon="fa-trash-alt"] { width: 0.875em; }
.svg-inline--fa[data-icon="fa-trash-can"] { width: 0.875em; }
